import unittest

from cert_core import Chain
from pycoin.encoding.hexbytes import b2h

from cert_issuer.merkle_tree_generator import MerkleTreeGenerator


def get_test_data_generator():
    """
    Returns a generator (1-time iterator) of test data
    :return:
    """
    for num in range(1, 4):
        yield str(num).encode('utf-8')


class TestMerkleTreeGenerator(unittest.TestCase):
    def test_generate(self):
        merkle_tree_generator = MerkleTreeGenerator()
        merkle_tree_generator.populate(get_test_data_generator())
        byte_array = merkle_tree_generator.get_blockchain_data()
        self.assertEqual(b2h(byte_array), '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044')

    def test_proofs_bitcoin_mainnet(self):
        self.do_test_signature(Chain.bitcoin_mainnet, 'bitcoinMainnet', 'BTCOpReturn')

    def test_proofs_bitcoin_testnet(self):
        self.do_test_signature(Chain.bitcoin_testnet, 'bitcoinTestnet', 'BTCOpReturn')

    def test_proofs_bitcoin_regtest(self):
        self.do_test_signature(Chain.bitcoin_regtest, 'bitcoinRegtest', 'BTCOpReturn')

    def test_proofs_mock(self):
        self.do_test_signature(Chain.mockchain, 'mockchain', 'Mock')

    def do_test_signature(self, chain, display_chain, type):
        merkle_tree_generator = MerkleTreeGenerator()
        merkle_tree_generator.populate(get_test_data_generator())
        _ = merkle_tree_generator.get_blockchain_data()
        gen = merkle_tree_generator.get_proof_generator(
            '8087c03e7b7bc9ca7b355de9d9d8165cc5c76307f337f0deb8a204d002c8e582', chain)
        p1 = next(gen)
        _ = next(gen)
        p3 = next(gen)
        p1_expected = {'type': ['MerkleProof2017', 'Extension'],
                       'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
                       'targetHash': '6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b',
                       'proof': [{'right': 'd4735e3a265e16eee03f59718b9b5d03019c07d8b6c51f90da3a666eec13ab35'},
                                 {'right': '4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce'}],
                       'anchors': [
                           {'sourceId': '8087c03e7b7bc9ca7b355de9d9d8165cc5c76307f337f0deb8a204d002c8e582',
                            'type': type,
                            'chain': display_chain}]}
        p3_expected = {'type': ['MerkleProof2017', 'Extension'],
                       'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
                       'targetHash': '4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce',
                       'proof': [{'left': '4295f72eeb1e3507b8461e240e3b8d18c1e7bd2f1122b11fc9ec40a65894031a'}],
                       'anchors': [{'sourceId': '8087c03e7b7bc9ca7b355de9d9d8165cc5c76307f337f0deb8a204d002c8e582',
                                    'type': type,
                                    'chain': display_chain}]}
        self.assertEqual(p1, p1_expected)
        self.assertEqual(p3, p3_expected)


if __name__ == '__main__':
    unittest.main()



class Error(Exception):
    """Base class for exceptions in this module"""
    pass


class InsufficientFundsError(Error):
    """
    There are insufficient funds to issue certificates
    """
    pass


class ConnectorError(Error):
    pass


class UnverifiedSignatureError(Error):
    """
    The signature in the certificate does not match the issuer's address
    """
    pass


class UnableToSignTxError(Error):
    """
    The transaction could not be signed
    """
    pass


class UnverifiedTransactionError(Error):
    """
    The transaction could not be verified
    """
    pass


class AlreadySignedError(Error):
    """
    The certificate has already been signed
    """
    pass


class NoCertificatesFoundError(Error):
    """
    No certificates found
    """
    pass


class NonemptyOutputDirectoryError(Error):
    """
    The output directory is not empty
    """
    pass


class BroadcastError(Error):
    """
    Error broadcasting transaction
    """
    pass


class UnrecognizedChainError(Error):
    """
    Didn't recognize chain
    """
    pass



import json
import logging

from bitcoin.signmessage import BitcoinMessage, SignMessage
from bitcoin.signmessage import VerifyMessage
from bitcoin.wallet import CBitcoinSecret
from pycoin.networks.registry import network_for_netcode
from pycoin.solve.utils import build_hash160_lookup

from cert_issuer.errors import UnverifiedSignatureError, UnableToSignTxError
from cert_issuer.helpers import to_pycoin_chain
from cert_issuer.models import Signer


class BitcoinSigner(Signer):
    def __init__(self, bitcoin_chain):
        self.bitcoin_chain = bitcoin_chain

    def sign_message(self, wif, message_to_sign):
        secret_key = CBitcoinSecret(wif)
        message = BitcoinMessage(message_to_sign)
        signature = SignMessage(secret_key, message)
        return str(signature, 'utf-8')

    def sign_transaction(self, wif, transaction_to_sign):
        netcode = to_pycoin_chain(self.bitcoin_chain)
        network = network_for_netcode(netcode)
        key = network.parse.wif(wif)
        secret_exponent = key.secret_exponent()
        lookup = build_hash160_lookup([secret_exponent], generators=[network.generator])
        signed_transaction = transaction_to_sign.sign(lookup)
        # Because signing failures silently continue, first check that the inputs are signed
        for input in signed_transaction.txs_in:
            if len(input.script) == 0:
                logging.error('Unable to sign transaction. hextx=%s', signed_transaction.as_hex())
                raise UnableToSignTxError('Unable to sign transaction')
        return signed_transaction


def verify_message(address, message, signature):
    """
    Verify message was signed by the address
    :param address: signing address
    :param message: message to check
    :param signature: signature being tested
    :return:
    """
    bitcoin_message = BitcoinMessage(message)
    verified = VerifyMessage(address, bitcoin_message, signature)
    return verified


def verify_signature(uid, signed_cert_file_name, issuing_address):
    """
    Verify the certificate signature matches the expected. Double-check the uid field in the certificate and use
    VerifyMessage to confirm that the signature in the certificate matches the issuing_address.

    Raises error is verification fails.

    Raises UnverifiedSignatureError if signature is invalid

    :param uid:
    :param signed_cert_file_name:
    :param issuing_address:
    :return:
    """

    logging.info('verifying signature for certificate with uid=%s:', uid)
    with open(signed_cert_file_name) as in_file:
        signed_cert = in_file.read()
        signed_cert_json = json.loads(signed_cert)
        to_verify = uid
        signature = signed_cert_json['signature']
        verified = verify_message(issuing_address, to_verify, signature)
        if not verified:
            error_message = 'There was a problem with the signature for certificate uid={}'.format(uid)
            raise UnverifiedSignatureError(error_message)

        logging.info('verified signature')



import hashlib

from cert_core import Chain
from blockcerts_merkletools import MerkleTools
from pycoin.encoding.hexbytes import h2b


def hash_byte_array(data):
    hashed = hashlib.sha256(data).hexdigest()
    return hashed


def ensure_string(value):
    if isinstance(value, str):
        return value
    return value.decode('utf-8')


class MerkleTreeGenerator(object):
    def __init__(self):
        self.tree = MerkleTools(hash_type='sha256')

    def populate(self, node_generator):
        """
        Populate Merkle Tree with data from node_generator. This requires that node_generator yield byte[] elements.
        Hashes, computes hex digest, and adds it to the Merkle Tree
        :param node_generator:
        :return:
        """
        for data in node_generator:
            hashed = hash_byte_array(data)
            self.tree.add_leaf(hashed)

    def get_blockchain_data(self):
        """
        Finalize tree and return byte array to issue on blockchain
        :return:
        """
        self.tree.make_tree()
        merkle_root = self.tree.get_merkle_root()
        return h2b(ensure_string(merkle_root))

    def get_proof_generator(self, tx_id, chain=Chain.bitcoin_mainnet):
        """
        Returns a generator (1-time iterator) of proofs in insertion order.

        :param tx_id: blockchain transaction id
        :return:
        """
        root = ensure_string(self.tree.get_merkle_root())
        node_count = len(self.tree.leaves)
        for index in range(0, node_count):
            proof = self.tree.get_proof(index)
            proof2 = []

            for p in proof:
                dict2 = dict()
                for key, value in p.items():
                    dict2[key] = ensure_string(value)
                proof2.append(dict2)
            target_hash = ensure_string(self.tree.get_leaf(index))
            merkle_proof = {
                "type": ['MerkleProof2017', 'Extension'],
                "merkleRoot": root,
                "targetHash": target_hash,
                "proof": proof2,
                "anchors": [{
                    "sourceId": to_source_id(tx_id, chain),
                    "type": chain.blockchain_type.external_display_value,
                    "chain": chain.external_display_value
                }]}
            yield merkle_proof


def to_source_id(txid, chain):
    if chain == Chain.bitcoin_mainnet or Chain.bitcoin_testnet or Chain.ethereum_mainnet or Chain.ethereum_ropsten:
        return txid
    else:
        return 'This has not been issued on a blockchain and is for testing only'



import unittest

import mock
import json
import io
from pycoin.encoding.hexbytes import b2h
from mock import patch, mock_open

from cert_issuer.certificate_handlers import CertificateWebV2Handler, CertificateV2Handler, CertificateBatchHandler, CertificateHandler, CertificateBatchWebHandler
from cert_issuer.merkle_tree_generator import MerkleTreeGenerator
from cert_issuer import helpers
from mock import ANY

class TestCertificateHandler(unittest.TestCase):
    def _proof_helper(self, chain):
        proof = {
                'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
                'type': ['MerkleProof2017', 'Extension'],
        'targetHash': ANY,
                'anchors': [
            {
                'sourceId': ANY,
                'type': chain.blockchain_type.external_display_value,
                'chain': chain.external_display_value
            }
                    ],
                'proof': [
                    {'right': 'd4735e3a265e16eee03f59718b9b5d03019c07d8b6c51f90da3a666eec13ab35'},
                    {'right': '4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce'}
                    ],
                }

        proof_1 = {
                'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
                'type': ['MerkleProof2017', 'Extension'],
                'targetHash': ANY,
                'anchors': [
                    {
                        'sourceId': ANY,
                        'type': chain.blockchain_type.external_display_value,
                        'chain': chain.external_display_value
                        }
                    ],
                'proof': [
                    {'left': '6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b'},
                    {'right': '4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce'}
                    ]
                }

        proof_2 = {
                'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
                'type': ['MerkleProof2017', 'Extension'],
                'targetHash': ANY,
                'anchors': [
                    {
                        'sourceId': ANY,
                        'type': chain.blockchain_type.external_display_value,
                        'chain': chain.external_display_value
                        }
                    ],
                'proof': [
                    {'left': '4295f72eeb1e3507b8461e240e3b8d18c1e7bd2f1122b11fc9ec40a65894031a'}
                    ]
                }
        return proof, proof_1, proof_2

    def _helper_mock_call(self, *args):
        helper_mock = mock.MagicMock()
        helper_mock.__len__.return_value = self.directory_count

        assert args == (
            '/unsigned_certificates_dir',
            '/signed_certificates_dir',
            '/blockchain_certificates_dir',
            '/work_dir')
        
        return helper_mock

    def _get_certificate_batch_web_handler(self):
        secret_manager = mock.Mock()
        certificates_to_issue = dict()
        certificates_to_issue['1'] = mock.Mock()
        certificates_to_issue['2'] = mock.Mock()
        certificates_to_issue['3'] = mock.Mock()

        handler = CertificateBatchWebHandler(
                secret_manager=secret_manager,
                certificate_handler=DummyCertificateHandler(),
                merkle_tree=MerkleTreeGenerator())

        return handler, certificates_to_issue

    def _get_certificate_batch_handler(self):
        secret_manager = mock.Mock()
        certificates_to_issue = dict()
        certificates_to_issue['1'] = mock.Mock()
        certificates_to_issue['2'] = mock.Mock()
        certificates_to_issue['3'] = mock.Mock()

        handler = CertificateBatchHandler(
                secret_manager=secret_manager,
                certificate_handler=DummyCertificateHandler(),
                merkle_tree=MerkleTreeGenerator())

        return handler, certificates_to_issue

    def test_batch_handler_web_prepare_batch(self):
        web_handler, certificates_to_issue = self._get_certificate_batch_web_handler()

        web_handler.set_certificates_in_batch(certificates_to_issue)
        result = web_handler.prepare_batch()
        self.assertEqual(
                b2h(result), '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044')

    def test_batch_handler_prepare_batch(self):
        certificate_batch_handler, certificates_to_issue = self._get_certificate_batch_handler()

        certificate_batch_handler.set_certificates_in_batch(certificates_to_issue)
        result = certificate_batch_handler.prepare_batch()
        self.assertEqual(
                b2h(result), '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044')

    def test_batch_web_handler_finish_batch(self):
        certificate_batch_handler, certificates_to_issue = self._get_certificate_batch_web_handler()

        certificate_batch_handler.set_certificates_in_batch(certificates_to_issue)
        result = certificate_batch_handler.prepare_batch()

        chain = mock.Mock()
        proof, proof_1, proof_2 = self._proof_helper(chain)

        with patch.object(DummyCertificateHandler, 'add_proof', return_value= {"cert": "cert"} ) as mock_method:
            result = certificate_batch_handler.finish_batch(
                        '5604f0c442922b5db54b69f8f363b3eac67835d36a006b98e8727f83b6a830c0', chain
                        )
        self.assertEqual(certificate_batch_handler.proof, [{'cert': 'cert'}, {'cert': 'cert'}, {'cert': 'cert'}])
        mock_method.assert_any_call(ANY, proof)
        mock_method.assert_any_call(ANY, proof_1)
        mock_method.assert_any_call(ANY, proof_2)

    def test_batch_handler_finish_batch(self):
        certificate_batch_handler, certificates_to_issue = self._get_certificate_batch_handler()

        certificate_batch_handler.set_certificates_in_batch(certificates_to_issue)
        result = certificate_batch_handler.prepare_batch()

        chain = mock.Mock()
        proof, proof_1, proof_2 = self._proof_helper(chain)

        with patch.object(DummyCertificateHandler, 'add_proof') as mock_method:
            result = certificate_batch_handler.finish_batch(
                    '5604f0c442922b5db54b69f8f363b3eac67835d36a006b98e8727f83b6a830c0', chain
                    )

        mock_method.assert_any_call(ANY, proof)
        mock_method.assert_any_call(ANY, proof_1)
        mock_method.assert_any_call(ANY, proof_2)

    def test_pre_batch_actions(self):
        self.directory_count = 1

        config = mock.Mock()
        config.unsigned_certificates_dir = '/unsigned_certificates_dir'
        config.signed_certificates_dir = '/signed_certificates_dir'
        config.blockchain_certificates_dir = '/blockchain_certificates_dir'
        config.work_dir = '/work_dir'

        helpers.prepare_issuance_batch = mock.MagicMock()

        helpers.prepare_issuance_batch.side_effect = self._helper_mock_call

        with patch.object(CertificateBatchHandler, 'set_certificates_in_batch') as mock_method:
            certificate_batch_handler, _ = self._get_certificate_batch_handler()
            certificate_batch_handler.pre_batch_actions(config)

        assert mock_method.called

    def test_pre_batch_actions_empty_directories(self):
        self.directory_count = 0

        config = mock.Mock()
        config.unsigned_certificates_dir = '/unsigned_certificates_dir'
        config.signed_certificates_dir = '/signed_certificates_dir'
        config.blockchain_certificates_dir = '/blockchain_certificates_dir'
        config.work_dir = '/work_dir'

        helpers.prepare_issuance_batch = mock.MagicMock()

        helpers.prepare_issuance_batch.side_effect = self._helper_mock_call

        with patch.object(CertificateBatchHandler, 'set_certificates_in_batch') as mock_method:
            certificate_batch_handler, _ = self._get_certificate_batch_handler()
            certificate_batch_handler.pre_batch_actions(config)

        assert not mock_method.called

    @mock.patch("builtins.open", create=True)
    def test_add_proof(self,mock_open):
        handler = CertificateV2Handler()

        cert_to_issue = {'kek':'kek'}
        proof = {'a': 'merkel'}
        file_call = 'call().__enter__().write(\'{"kek": "kek", "signature": {"a": "merkel"}}\')'

        chain = mock.Mock()
        metadata = mock.Mock()
        metadata.blockchain_cert_file_name = 'file_path.nfo'

        with patch.object(
        CertificateV2Handler, '_get_certificate_to_issue', return_value=cert_to_issue) as mock_method:
                handler.add_proof(metadata, proof)

        mock_open.assert_any_call('file_path.nfo','w')
        calls = mock_open.mock_calls
        call_strings = map(str, calls)
        assert file_call in call_strings

    def test_web_add_proof(self):
        handler = CertificateWebV2Handler()
        proof = {'a': 'merkel'}
        chain = mock.Mock()
        certificate_json = {'kek': 'kek'}

        return_cert = handler.add_proof(certificate_json, proof)
        self.assertEqual(return_cert, {'kek':'kek', 'signature': {'a': 'merkel'}})

class DummyCertificateHandler(CertificateHandler):
    def __init__(self):
        self.counter = 0

    def validate_certificate(self, certificate_metadata):
        pass

    def sign_certificate(self, signer, certificate_metadata):
        pass

    def get_byte_array_to_issue(self, certificate_metadata):
        self.counter += 1
        return str(self.counter).encode('utf-8')

    def add_proof(self, certificate_metadata, merkle_proof):
        pass


if __name__ == '__main__':
    unittest.main()



import logging
from web3 import Web3
from cert_issuer.errors import UnverifiedTransactionError


def create_ethereum_trx(issuing_address, nonce, to_address, blockchain_bytes, gasprice, gaslimit):
    # the actual value transfer is 0 in the Ethereum implementation
    value = 0
    tx = dict(
        nonce=nonce,
        gasPrice=gasprice,
        gas=gaslimit,
        to=Web3.to_checksum_address(to_address),
        value=value,
        data=blockchain_bytes)
    return tx


def verify_eth_transaction(signed_hextx, eth_data_field):
    """
    Verify ethDataField field in transaction
    :param signed_hextx:
    :param eth_data_field:
    :return:
    """
    logging.info('verifying ethDataField value for transaction')
    ethdata_hash = []
    for s in signed_hextx.split('80a0'):
        ethdata_hash.append(s)
    ethdata_hash = ethdata_hash[1][:64]
    result = (eth_data_field == ethdata_hash)
    if not result:
        error_message = 'There was a problem verifying the transaction'
        raise UnverifiedTransactionError(error_message)
    logging.info('verified ethDataField')



import logging
import time

import requests
import web3
from web3 import Web3, HTTPProvider

try:
    from urllib2 import urlopen, HTTPError
    from urllib import urlencode
except ImportError:
    from urllib.request import urlopen, HTTPError
    from urllib.parse import urlencode

from cert_core import Chain
from cert_issuer.models import ServiceProviderConnector
from cert_issuer.errors import BroadcastError

POST_BROADCAST_DELAY_SECONDS = 2
BROADCAST_RETRY_INTERVAL = 30
MAX_BROADCAST_ATTEMPTS = 3


class EthereumServiceProviderConnector(ServiceProviderConnector):
    # param local_node indicates if a local node is running or if the tx should be broadcast to external providers
    def __init__(
            self,
            ethereum_chain,
            app_config,
            local_node=False):
        self.ethereum_chain = ethereum_chain

        self.local_node = local_node

        # initialize connectors
        self.connectors = {}

        # Configure Ethereum mainnet connectors
        eth_provider_list = []
        if hasattr(app_config, 'ethereum_rpc_url'):
            self.ethereum_rpc_url = app_config.ethereum_rpc_url
            eth_provider_list.append(EthereumRPCProvider(self.ethereum_rpc_url))

        etherscan_api_token = None
        if hasattr(app_config, 'api_token'):
            logging.warning('The api_token config property is deprecated in favor of the etherscan_api_token property.  It still works, but please switch over soon.')
            etherscan_api_token = app_config.etherscan_api_token
        if hasattr(app_config, 'etherscan_api_token'):
            etherscan_api_token = app_config.etherscan_api_token
        eth_provider_list.append(EtherscanBroadcaster(self.get_etherscan_api_endpoint_for_chain('main'), etherscan_api_token))
        # eth_provider_list.append(MyEtherWalletBroadcaster('https://api.myetherwallet.com/eth', None))
        self.connectors[Chain.ethereum_mainnet] = eth_provider_list

        # Configure Ethereum Ropsten testnet connectors
        rop_provider_list = []
        if hasattr(app_config, 'ropsten_rpc_url'):
            self.ropsten_rpc_url = app_config.ropsten_rpc_url
            rop_provider_list.append(EthereumRPCProvider(self.ropsten_rpc_url))
            rop_provider_list.append(EtherscanBroadcaster('https://api-ropsten.etherscan.io/api', etherscan_api_token))
        # rop_provider_list.append(MyEtherWalletBroadcaster('https://api.myetherwallet.com/rop', None))
        self.connectors[Chain.ethereum_ropsten] = rop_provider_list

    def get_etherscan_api_endpoint_for_chain(self, chain):
        chain_codes = {
            'main': 1,
            'goerli': 5,
            'sepolia': 11155111
        }

        chainId = chain_codes[chain]

        if chainId is None:
            raise ValueError(f'Unknown chain code for Etherscan API: {chain}')

        logging.debug(f'Etherscan endpoint url for chain {chain} is https://api.etherscan.io/v2/api?chainid={chainId}')

        return f'https://api.etherscan.io/v2/api?chainid={chainId}'

    def get_providers_for_chain(self, chain, local_node=False):
        return self.connectors[chain]

    def get_balance(self, address):
        for m in self.get_providers_for_chain(self.ethereum_chain, self.local_node):
            try:
                logging.debug('m=%s', m)
                balance = m.get_balance(address)
                return balance
            except Exception as e:
                logging.warning(e)
                pass
        return 0

    def get_address_nonce(self, address):
        for m in self.get_providers_for_chain(self.ethereum_chain, self.local_node):
            try:
                logging.debug('m=%s', m)
                nonce = m.get_address_nonce(address)
                return nonce
            except Exception as e:
                logging.warning(e)
                pass
        return 0
    
    def check_tx_exists(self, txid):
        for m in self.get_providers_for_chain(self.ethereum_chain, self.local_node):
            try:
                logging.debug('m=%s', m)
                if hasattr(m, 'tx_seen_in_mempool') and m.tx_seen_in_mempool(txid):
                    return True
            except Exception as e:
                logging.warning('Provider check for tx_seen_in_mempool failed: %s', e)
        return False

    def broadcast_tx(self, tx):

        last_exception = None
        final_tx_id = None

        # Broadcast to all available api's
        for attempt_number in range(0, MAX_BROADCAST_ATTEMPTS):
            for m in self.get_providers_for_chain(self.ethereum_chain, self.local_node):
                try:
                    logging.debug('m=%s', m)
                    txid = m.broadcast_tx(tx)
                    if (txid):
                        logging.info('Broadcasting succeeded with method_provider=%s, txid=%s', str(m), txid)
                        if final_tx_id and final_tx_id != txid:
                            logging.error(
                                'This should never happen; fail and investigate if it does. Got conflicting tx_ids=%s and %s. Hextx=%s',
                                final_tx_id, txid, tx.as_hex())
                            raise Exception('Got conflicting tx_ids.')
                        final_tx_id = txid

                    if final_tx_id:
                        # Wait a couple of seconds, then check the network to see if the transaction made it.
                        time.sleep(POST_BROADCAST_DELAY_SECONDS)
                        if not m.tx_seen_in_mempool(final_tx_id):
                            logging.warning("Transaction %s not present in mempool or chain!", final_tx_id)
                        else:
                            logging.info("Transaction %s confirmed visible to at least one provider", final_tx_id)

                    return txid

                except Exception as e:
                    logging.warning('Caught exception trying provider %s. Trying another. Exception=%s',
                                    str(m), e)
                    last_exception = e

            # At least 1 provider succeeded, so return
            if final_tx_id:
                return final_tx_id
            else:
                logging.warning('Broadcasting failed. Waiting before retrying. This is attempt number %d',
                                attempt_number)
                time.sleep(BROADCAST_RETRY_INTERVAL)

        ##in case of failure:
        logging.error('Failed broadcasting through all providers')
        logging.error(last_exception, exc_info=True)
        raise BroadcastError(last_exception)


class EthereumRPCProvider(object):
    def __init__(self, ethereum_url):
        self.ethereum_url = ethereum_url
        self.w3 = Web3(HTTPProvider(ethereum_url))

    def broadcast_tx(self, tx):
        logging.info('Broadcasting transaction with EthereumRPCProvider')
        response = self.w3.eth.send_raw_transaction("0x" + tx).hex()
        return response

    def get_balance(self, address):
        """
        Returns the balance in Wei.
        """
        logging.info('Getting balance with EthereumRPCProvider')
        response = self.w3.eth.get_balance(account=address, block_identifier="latest")
        return response

    def get_address_nonce(self, address):
        """
        Looks up the address nonce of this address.
        Necessary for the transaction creation.
        """
        logging.info('Fetching nonce with EthereumRPCProvider')
        response = self.w3.eth.get_transaction_count(address, "pending")
        return response
    
    def tx_seen_in_mempool(self, txid):
        try:
            result = self.w3.eth.get_transaction(txid)
            return result is not None
        except Exception:
            return False


class EtherscanBroadcaster(object):
    def __init__(self, base_url, api_token):
        self.base_url = base_url
        self.api_token = api_token

    def broadcast_tx(self, tx):
        tx_hex = tx

        broadcast_url = self.base_url + '&module=proxy&action=eth_sendRawTransaction'
        if self.api_token:
            broadcast_url += '&apikey=%s' % self.api_token
        response = requests.post(broadcast_url, data={'hex': tx_hex}, headers={'user-agent':'cert-issuer'})

        if 'error' in response.json():
            logging.error("Etherscan returned an error: %s", response.json()['error'])
            raise BroadcastError(response.json()['error'])
        if int(response.status_code) == 200:
            if response.json().get('message', None) == 'NOTOK':
                raise BroadcastError(response.json().get('result', None))
            tx_id = response.json().get('result', None)
            logging.info("Transaction ID obtained from broadcast through Etherscan: %s", tx_id)
            return tx_id
        logging.error('Error broadcasting the transaction through the Etherscan API. Error msg: %s', response.text)
        raise BroadcastError(response.text)

    def get_balance(self, address):
        """
        returns the balance in wei
        with some inspiration from PyWallet
        """
        broadcast_url = self.base_url + '&module=account&action=balance'
        broadcast_url += '&address=%s' % address
        broadcast_url += '&tag=pending'
        if self.api_token:
            broadcast_url += '&apikey=%s' % self.api_token
        response = requests.get(broadcast_url, headers={'user-agent':'cert-issuer'})
        if int(response.status_code) == 200:
            if response.json().get('message', None) == 'NOTOK':
                raise BroadcastError(response.json().get('result', None))
            balance = int(response.json().get('result', None))
            logging.info('Balance check succeeded: %s', response.json())
            return balance
        raise BroadcastError(response.text)

    def get_address_nonce(self, address):
        """
        Looks up the address nonce of this address
        Neccesary for the transaction creation
        """
        broadcast_url = self.base_url + '&module=proxy&action=eth_getTransactionCount'
        broadcast_url += '&address=%s' % address
        broadcast_url += '&tag=pending' # Valid tags are 'earliest', 'latest', and 'pending', the last of which includes both pending and committed transactions.
        if self.api_token:
            broadcast_url += '&apikey=%s' % self.api_token
        response = requests.get(broadcast_url, headers={'user-agent':'cert-issuer'})
        if int(response.status_code) == 200:
            if response.json().get('message', None) == 'NOTOK':
                raise BroadcastError(response.json().get('result', None))
            nonce = int(response.json().get('result', None), 0)
            logging.info('Nonce check went correct: %s', response.json())
            return nonce
        else:
            logging.info('response error checking nonce')
        raise BroadcastError('Error checking the nonce through the Etherscan API. Error msg: %s', response.text)
    
    def tx_seen_in_mempool(self, txid):
        url = self.base_url + '&module=proxy&action=eth_getTransactionByHash' \
              + f'&txhash={txid}'
        if self.api_token:
            url += f'&apikey={self.api_token}'
        response = requests.get(url)
        if int(response.status_code) != 200:
            return False
        return response.json().get("result") is not None


class MyEtherWalletBroadcaster(object):
    def __init__(self, base_url, api_token):
        self.base_url = base_url
        self.api_token = api_token

    def broadcast_tx(self, tx):
        data = {
            "jsonrpc": "2.0",
            "method": "eth_sendRawTransaction",
            "params": ["0x" + tx],
            "id": 1
        }
        response = requests.post(self.base_url, json=data, headers={'user-agent':'cert-issuer'})
        if 'error' in response.json():
            logging.error("MyEtherWallet returned an error: %s", response.json()['error'])
            raise BroadcastError(response.json()['error'])
        if int(response.status_code) == 200:
            tx_id = response.json().get('result', None)
            logging.info("Transaction ID obtained from broadcast through MyEtherWallet: %s", tx_id)
            return tx_id
        logging.error('Error broadcasting the transaction through MyEtherWallet. Error msg: %s', response.text)
        raise BroadcastError(response.text)

    def get_balance(self, address):
        """
        returns the balance in wei
        """

        data = {
            "jsonrpc": "2.0",
            "method": "eth_getBalance",
            "params": [address, "latest"],
            "id": 1
        }
        response = requests.post(self.base_url, json=data, headers={'user-agent':'cert-issuer'})
        if int(response.status_code) == 200:
            logging.info('Balance check response: %s', response.json())
            balance = int(response.json().get('result', None), 0)
            logging.info('Balance check succeeded: %s', response.json())
            return balance
        logging.error('Error getting balance through MyEtherWallet. Error msg: %s', response.text)
        raise BroadcastError(response.text)

    def get_address_nonce(self, address):
        """
        Looks up the address nonce of this address
        Neccesary for the transaction creation
        """

        data = {
            "jsonrpc": "2.0",
            "method": "eth_getTransactionCount",
            "params": [address, "pending"],
            "id": 1
        }
        response = requests.post(self.base_url, json=data, headers={'user-agent':'cert-issuer'})
        if int(response.status_code) == 200:
            # the int(res, 0) transforms the hex nonce to int
            nonce = int(response.json().get('result', None), 0)
            logging.info('Nonce check went correct: %s', response.json())
            return nonce
        else:
            logging.info('response error checking nonce')
        raise BroadcastError('Error checking the nonce through the MyEtherWallet API. Error msg: %s', response.text)



import logging
import os

from cert_core import BlockchainType
from cert_core import Chain, UnknownChainError

from cert_issuer.certificate_handlers import CertificateBatchHandler, CertificateV2Handler
from cert_issuer.blockchain_handlers.ethereum.connectors import EthereumServiceProviderConnector
from cert_issuer.blockchain_handlers.ethereum.signer import EthereumSigner
from cert_issuer.blockchain_handlers.ethereum.transaction_handlers import EthereumTransactionHandler
from cert_issuer.merkle_tree_generator import MerkleTreeGenerator
from cert_issuer.models import MockTransactionHandler
from cert_issuer.signer import FileSecretManager


class EthereumTransactionCostConstants(object):
    def __init__(self, recommended_gas_price=20000000000, recommended_gas_limit=25000):
        self.recommended_gas_price = recommended_gas_price
        self.recommended_gas_limit = recommended_gas_limit
        logging.info('Set cost constants to recommended_gas_price=%f, recommended_gas_limit=%f',
                     self.recommended_gas_price, self.recommended_gas_limit)

    """
    The below methods currently only use the supplied gasprice/limit.
    These values can also be better estimated via a call to the Ethereum blockchain.
    """

    def get_recommended_max_cost(self):
        return self.recommended_gas_price * self.recommended_gas_limit

    def get_gas_price(self):
        return self.recommended_gas_price

    def get_gas_limit(self):
        return self.recommended_gas_limit


def initialize_signer(app_config):
    path_to_secret = os.path.join(app_config.usb_name, app_config.key_file)

    if app_config.chain.blockchain_type == BlockchainType.ethereum:
        signer = EthereumSigner(ethereum_chain=app_config.chain)
    elif app_config.chain == Chain.mockchain:
        signer = None
    else:
        raise UnknownChainError(app_config.chain)
    secret_manager = FileSecretManager(signer=signer, path_to_secret=path_to_secret,
                                       safe_mode=app_config.safe_mode, issuing_address=app_config.issuing_address)
    return secret_manager


def instantiate_blockchain_handlers(app_config):
    issuing_address = app_config.issuing_address
    chain = app_config.chain
    secret_manager = initialize_signer(app_config)
    certificate_batch_handler = CertificateBatchHandler(secret_manager=secret_manager,
                                                        certificate_handler=CertificateV2Handler(),
                                                        merkle_tree=MerkleTreeGenerator())
    if chain == Chain.mockchain:
        transaction_handler = MockTransactionHandler()
    # ethereum chains
    elif chain == Chain.ethereum_mainnet or chain == Chain.ethereum_ropsten:
        cost_constants = EthereumTransactionCostConstants(app_config.gas_price, app_config.gas_limit)
        connector = EthereumServiceProviderConnector(chain, app_config)
        transaction_handler = EthereumTransactionHandler(connector, cost_constants, secret_manager,
                                                         issuing_address=issuing_address)

    return certificate_batch_handler, transaction_handler, connector
